@@ -1,4 +1,7 @@
+#include <algorithm>
 #include <cassert>
+#include <cstdlib>
+#include <iterator>
 #include <thread>
 
 #include <pkgxx/makevars.hxx>
@@ -27,10 +30,9 @@ namespace pkg_chk {
             std::async(
                 std::launch::deferred,
                 [this]() {
-                    atomic_verbose(
-                        [](auto& out) {
-                            out << "Getting summary from installed packages" << std::endl;
-                        });
+                    verbose([](auto& out) {
+                        out << "Getting summary from installed packages" << std::endl;
+                    });
                     return pkgxx::summary(_PKG_INFO.get());
                 }).share())
         , _installed_pkgnames(
@@ -42,20 +44,12 @@ namespace pkg_chk {
                         ret.insert(name);
                     }
                     return ret;
-                }).share())
-        , _installed_pkgpaths(
-            std::async(
-                std::launch::deferred,
-                [this]() {
-                    std::set<pkgxx::pkgpath> ret;
-                    for (auto const& [_name, vars]: _installed_pkg_summary.get()) {
-                        ret.insert(vars.PKGPATH);
-                    }
-                    return ret;
                 }).share()) {}
 
     checker_base::result
     checker_base::run(std::set<pkgxx::pkgpath> const& pkgpaths) const {
+        total(pkgpaths.size());
+
         // This is the slowest part of pkg_chk. For each package we need to
         // extract variables from package Makefiles unless we are using
         // binary packages. Luckily for us each check is independent of
@@ -72,6 +66,7 @@ namespace pkg_chk {
                         auto const latest_pkgnames = find_latest_pkgnames(path);
                         if (latest_pkgnames.empty()) {
                             res.lock()->MISSING_DONE.insert(path);
+                            progress();
                             return;
                         }
 
@@ -79,7 +74,9 @@ namespace pkg_chk {
                         for (pkgxx::pkgname const& name: latest_pkgnames) {
                             if (auto installed = installed_pkgnames.lower_bound(
                                     pkgxx::pkgname(name.base, pkgxx::pkgversion()));
-                                installed != installed_pkgnames.end() && installed->base == name.base) {
+                                installed != installed_pkgnames.end() &&
+                                installed->base == name.base &&
+                                !_deleted_pkgnames.count(*installed)) {
 
                                 if (installed->version == name.version) {
                                     // The latest PKGNAME turned out to be
@@ -93,87 +90,81 @@ namespace pkg_chk {
 
                                         if (latest_build_version.has_value()) {
                                             if (latest_build_version == installed_build_version) {
-                                                atomic_verbose(
-                                                    [&](auto& out) {
-                                                        out << path << " - " << name << " OK" << std::endl;
-                                                    });
+                                                verbose([&](auto& out) {
+                                                    out << path << " - " << name << " OK" << std::endl;
+                                                });
                                             }
                                             else {
-                                                atomic_msg(
-                                                    [&](auto& out) {
-                                                        out << path << " - " << name << " build_version mismatch" << std::endl;
-                                                    });
-                                                atomic_verbose(
-                                                    [&](auto& out) {
-                                                        out << "--current--"                   << std::endl
-                                                            << latest_build_version.value()
-                                                            << "--installed--"                 << std::endl
-                                                            << installed_build_version.value()
-                                                            << "----"                          << std::endl
-                                                            << std::endl;
-                                                    });
+                                                msg([&](auto& out) {
+                                                    out << path << " - " << name << " build_version mismatch" << std::endl;
+                                                });
+                                                verbose([&](auto& out) {
+                                                    out << "--current--"                   << std::endl
+                                                        << latest_build_version.value()
+                                                        << "--installed--"                 << std::endl
+                                                        << installed_build_version.value()
+                                                        << "----"                          << std::endl
+                                                        << std::endl;
+                                                });
                                                 res.lock()->MISMATCH_TODO.emplace(*installed, path);
                                             }
                                         }
                                         else {
-                                            atomic_msg(
-                                                [&](auto& out) {
-                                                    out << path << " - " << name << " build_version missing" << std::endl;
-                                                });
+                                            msg([&](auto& out) {
+                                                out << path << " - " << name << " build_version missing" << std::endl;
+                                            });
                                         }
                                     }
                                     else {
-                                        atomic_verbose(
-                                            [&](auto& out) {
-                                                out << path << " - " << name << " OK" << std::endl;
-                                            });
+                                        verbose([&](auto& out) {
+                                            out << path << " - " << name << " OK" << std::endl;
+                                        });
                                     }
                                 }
                                 else if (installed->version < name.version) {
                                     // We have an older version installed.
-                                    atomic_msg(
-                                        [&](auto& out) {
-                                            out << path << " - " << *installed << " < " << name
-                                                << (is_binary_available(name) ? " (has binary package)" : "")
-                                                << std::endl;
-                                        });
+                                    msg([&](auto& out) {
+                                        out << path << " - " << *installed << " < " << name
+                                            << (is_binary_available(name) ? " (has binary package)" : "")
+                                            << std::endl;
+                                    });
                                     res.lock()->MISMATCH_TODO.emplace(*installed, path);
                                 }
                                 else {
                                     // We have a newer version installed
                                     // but how can that happen?
                                     if (_check_build_version) {
-                                        atomic_msg(
-                                            [&](auto& out) {
-                                                out << path << " - " << *installed << " > " << name
-                                                    << (is_binary_available(name) ? " (has binary package)" : "")
-                                                    << std::endl;
-                                            });
+                                        msg([&](auto& out) {
+                                            out << path << " - " << *installed << " > " << name
+                                                << (is_binary_available(name) ? " (has binary package)" : "")
+                                                << std::endl;
+                                        });
                                         res.lock()->MISMATCH_TODO.emplace(*installed, path);
                                     }
                                     else {
-                                        atomic_msg(
-                                            [&](auto& out) {
-                                                out << path << " - " << *installed << " > " << name << " - ignoring"
-                                                    << (is_binary_available(name) ? " (has binary package)" : "")
-                                                    << std::endl;
-                                            });
+                                        msg([&](auto& out) {
+                                            out << path << " - " << *installed << " > " << name << " - ignoring"
+                                                << (is_binary_available(name) ? " (has binary package)" : "")
+                                                << std::endl;
+                                        });
                                     }
                                 }
                             }
                             else {
-                                atomic_msg(
-                                    [&](auto& out) {
-                                        out << path << " - " << name << " missing"
-                                            << (is_binary_available(name) ? " (has binary package)" : "")
-                                            << std::endl;
-                                    });
+                                msg([&](auto& out) {
+                                    out << path << " - " << name << " missing"
+                                        << (is_binary_available(name) ? " (has binary package)" : "")
+                                        << std::endl;
+                                });
                                 res.lock()->MISSING_TODO.emplace(name, path);
                             }
                         }
+                        progress();
                     });
             }
         }
+        done();
+
         // The nursery has to be destroyed before this std::move() happens,
         // otherwise we would return an empty result.
         return std::move(*res.lock());
@@ -181,7 +172,19 @@ namespace pkg_chk {
 
     checker_base::result
     checker_base::run() const {
-        return run(_installed_pkgpaths.get());
+        std::set<pkgxx::pkgpath> pkgpaths;
+        for (auto const& [name, vars]: _installed_pkg_summary.get()) {
+            if (!_deleted_pkgnames.count(name)) {
+                pkgpaths.insert(vars.PKGPATH);
+            }
+        }
+        return run(pkgpaths);
+    }
+
+    bool
+    checker_base::mark_as_deleted(pkgxx::pkgname const& name) {
+        auto const& [_, inserted] = _deleted_pkgnames.insert(name);
+        return inserted;
     }
 
     source_checker_base::source_checker_base(
@@ -216,20 +219,19 @@ namespace pkg_chk {
         // * pkg_chk -r: Same as above.
         //
         if (!fs::exists(_PKGSRCDIR.get() / path / "Makefile")) {
-            atomic_warn(
-                [&](auto& out) {
-                    out << "No " << path << "/Makefile - package moved or obsolete?" << std::endl;
-                });
+            warn([&](auto& out) {
+                out << "No " << path << "/Makefile - package moved or obsolete?" << std::endl;
+            });
             return {};
         }
 
         auto const default_pkgname =
             pkgxx::extract_pkgmk_var<pkgxx::pkgname>(_PKGSRCDIR.get() / path, "PKGNAME");
         if (!default_pkgname) {
-            fatal(
-                [&](auto& out) {
-                    out << "Unable to extract PKGNAME for " << path << std::endl;
-                });
+            fatal([&](auto& out) {
+                out << "Unable to extract PKGNAME for " << path << std::endl;
+            });
+            std::terminate(); // Should not reach here.
         }
 
         // We search non-default PKGNAMEs only when -u or -r is given,
@@ -242,7 +244,10 @@ namespace pkg_chk {
             auto const& pm = _installed_pkgpaths_with_pkgnames.get();
             if (auto installed_pkgnames = pm.find(path); installed_pkgnames != pm.end()) {
                 for (auto const& installed_pkgname: installed_pkgnames->second) {
-                    if (installed_pkgname.base != default_pkgname->base) {
+                    if (_deleted_pkgnames.count(installed_pkgname)) {
+                        continue;
+                    }
+                    else if (installed_pkgname.base != default_pkgname->base) {
                         // We found a non-default PKGBASE but spawning
                         // make(1) takes seriously long. It's really
                         // tempting to cheat by making up a PKGNAME by
@@ -264,14 +269,13 @@ namespace pkg_chk {
                             pkgnames.insert(std::move(alternative_pkgname));
                         }
                         else {
-                            atomic_warn(
-                                [&](auto& out) {
-                                    out << path << " had presumably provided a package named like "
-                                        << installed_pkgname.base << "-[0-9]* but it no longer does so. "
-                                        << "The installed package " << installed_pkgname
-                                        << " cannot be updated. Delete it and re-run the command."
-                                        << std::endl;
-                                });
+                            warn([&](auto& out) {
+                                out << path << " had presumably provided a package named like "
+                                    << installed_pkgname.base << "-[0-9]* but it no longer does so. "
+                                    << "The installed package " << installed_pkgname
+                                    << " cannot be updated. Delete it and re-run the command."
+                                    << std::endl;
+                            });
                             return {};
                         }
                     }
@@ -299,16 +303,6 @@ namespace pkg_chk {
                 std::launch::deferred,
                 [this]() {
                     return pkgxx::pkgmap(_bin_pkg_summary.get());
-                }).share())
-        , _installed_pkgbases(
-            std::async(
-                std::launch::deferred,
-                [this]() {
-                    std::set<pkgxx::pkgbase> ret;
-                    for (auto const& [name, _vars]: _installed_pkg_summary.get()) {
-                        ret.insert(name.base);
-                    }
-                    return ret;
                 }).share()) {}
 
     std::set<pkgxx::pkgname>
@@ -336,10 +330,15 @@ namespace pkg_chk {
             if (_update || _delete_mismatched) {
                 // We need to enumerate only PKGBASEs that are already
                 // installed, otherwise -a would install every single
-                // PKGNAME that the PKGPATH provides.
-                auto const& installed_pkgbases = _installed_pkgbases.get();
+                // package that the PKGPATH provides.
+                auto const& installed_pkgnames = _installed_pkgnames.get();
                 for (auto const& [base, sum]: pkgbases->second) {
-                    if (installed_pkgbases.find(base) != installed_pkgbases.end()) {
+                    if (auto installed = installed_pkgnames.lower_bound(
+                            pkgxx::pkgname(base, pkgxx::pkgversion()));
+                        installed != installed_pkgnames.end() &&
+                        installed->base == base &&
+                        !_deleted_pkgnames.count(*installed)) {
+
                         auto latest = sum.rbegin();
                         assert(latest != sum.rend());
 
