@@ -1,15 +1,19 @@
 #include <exception>
 #include <filesystem>
 #include <fstream>
+#include <memory>
 
 #include <pkgxx/config.h>
+#include <pkgxx/progress_bar.hxx>
 #include <pkgxx/string_algo.hxx>
 
 #include "pkg_chk/check.hxx"
 #include "replacer.hxx"
 
 using namespace std::chrono_literals;
+using namespace pkgxx::tty::literals;
 namespace fs = std::filesystem;
+namespace tty = pkgxx::tty;
 
 namespace {
     struct replace_failed: virtual std::runtime_error {
@@ -17,39 +21,60 @@ namespace {
     };
 
     struct source_checker: virtual pkg_chk::source_checker_base {
-        source_checker(pkg_rr::options const& opts, pkg_rr::environment const& env)
+        source_checker(pkg_rr::environment const& env)
             : checker_base(
                 false, // add_missing (-a)
-                opts.check_build_version,
-                opts.concurrency,
+                env.opts.check_build_version,
+                env.opts.concurrency,
                 true,  // update (-u)
                 false, // delete_mismatched (-r)
                 env.PKG_INFO)
             , source_checker_base(env.PKGSRCDIR)
-            , _opts(opts) {}
+            , _env(env) {}
 
     protected:
-        virtual void
-        atomic_msg(std::function<void (std::ostream&)> const& f) const override {
-            pkg_rr::atomic_msg(f);
+        virtual void total(std::size_t num) const override {
+            _pb = std::make_unique<pkgxx::progress_bar>(num);
+        }
+
+        virtual void progress() const override {
+            (*_pb)++;
+        }
+
+        virtual void done() const override {
+            _pb.reset();
         }
 
         virtual void
-        atomic_warn(std::function<void (std::ostream&)> const& f) const override {
-            pkg_rr::atomic_warn(f);
+        msg(std::function<void (pkgxx::ttystream_base&)> const& f) const override {
+            _pb->message([&](auto& out) {
+                auto out_ = _env.msg(out);
+                f(out_);
+            });
         }
 
         virtual void
-        atomic_verbose(std::function<void (std::ostream&)> const&) const override {
+        warn(std::function<void (pkgxx::ttystream_base&)> const& f) const override {
+            _pb->message([&](auto& out) {
+                auto out_ = _env.warn(out);
+                f(out_);
+            });
+        }
+
+        virtual void
+        verbose(std::function<void (pkgxx::ttystream_base&)> const&) const override {
             // Don't show verbose messages from pkg_chk.
         }
 
         virtual void
-        fatal(std::function<void (std::ostream&)> const& f) const override {
-            pkg_rr::fatal(f);
+        fatal(std::function<void (pkgxx::ttystream_base&)> const& f) const override {
+            _pb->message([&](auto& out) {
+                _env.fatal(f, out);
+            });
         }
 
-        pkg_rr::options const& _opts;
+        pkg_rr::environment _env;
+        mutable std::unique_ptr<pkgxx::progress_bar> _pb;
     };
 
     std::optional<pkgxx::pkgbase>
@@ -76,7 +101,12 @@ namespace pkg_rr {
         , opts(opts_)
         , env(env_)
         , UNSAFE_VAR(opts.strict ? "unsafe_depends_strict" : "unsafe_depends")
-        , pattern_to_base_cache(0) {
+        , pattern_to_base_cache(0)
+        , _pkgname_sty(tty::bold)
+        , _new_deps_sty(tty::faint)
+        , _var_sty(tty::faint)
+        , _even_sty(tty::dull_colour(tty::magenta))
+        , _odd_sty(tty::dull_colour(tty::cyan)) {
 
         std::future<todo_type> MISMATCH_TODO_f;
         std::future<todo_type> REBUILD_TODO_f;
@@ -112,10 +142,10 @@ namespace pkg_rr {
             auto [base, path] = choose_one();
             try {
                 if (DEPENDS_CHECKED.count(base)) {
-                    msg() << "Selecting " << base << " ("
-                          << static_cast<std::filesystem::path const&>(path).string()
-                          << ") as next package to replace" << std::endl;
-                    vsleep(opts, 1s);
+                    env.msg() << "Selecting " << _pkgname_sty(base) << " ("
+                              << static_cast<std::filesystem::path const&>(path).string()
+                              << ") as next package to replace" << std::endl;
+                    env.vsleep(1s);
                 }
                 else {
                     auto const version = update_depends_with_source(base, path);
@@ -133,12 +163,14 @@ namespace pkg_rr {
             }
             catch (replace_failed const& e) {
                 FAILED.push_back(base);
-                auto const& cb = [&](auto& out) { out << e.what() << std::endl; };
+
                 if (opts.continue_on_errors) {
-                    atomic_error(cb);
+                    env.error() << e.what() << std::endl;
                 }
                 else {
-                    abort(cb);
+                    abort([&](auto& out) {
+                        out << e.what() << std::endl;
+                    });
                 }
             }
 
@@ -151,22 +183,22 @@ namespace pkg_rr {
 
             refresh_todo();
             dump_todo();
-            vsleep(opts, 2s);
+            env.vsleep(2s);
         }
-        msg() << "No more packages to replace; done." << std::endl;
+        env.msg() << "No more packages to replace; done." << std::endl;
         report();
     }
 
     std::future<rolling_replacer::todo_type>
     rolling_replacer::check_mismatch(pkg_rr::package_scanner& scanner) const {
         if (opts.check_for_updates) {
-            msg() << "Checking for mismatched installed packages by scanning source tree" << std::endl;
-            auto result = source_checker(opts, env).run();
+            env.msg() << "Checking for mismatched installed packages by scanning source tree" << std::endl;
+            auto result = source_checker(env).run();
 
             if (!result.MISMATCH_TODO.empty()) {
                 // Spawn xargs only if it's non-empty; otherwise we would
                 // end up asking for password for nothing.
-                msg() << "Marking outdated packages as mismatched" << std::endl;
+                env.msg() << "Marking outdated packages as mismatched" << std::endl;
 
                 pkgxx::harness xargs =
                     spawn_su(std::string(CFG_XARGS) + ' ' + env.PKG_ADMIN.get() + " set mismatch=YES");
@@ -176,8 +208,8 @@ namespace pkg_rr {
                 xargs.cin().close();
 
                 if (xargs.wait_exit().status != 0) {
-                    warn() << "mismatch variable not set due to permissions; "
-                           << "the status will not persist." << std::endl;
+                    env.warn() << "mismatch variable not set due to permissions; "
+                               << "the status will not persist." << std::endl;
                 }
             }
 
@@ -192,7 +224,8 @@ namespace pkg_rr {
             return res.get_future();
         }
         else {
-            msg() << "Checking for mismatched installed packages (mismatch=YES)" << std::endl;
+            env.msg() << "Checking for mismatched installed packages "
+                      << _var_sty("(mismatch=YES)") << std::endl;
             return scanner.add_axis("mismatch", opts.no_check);
         }
     }
@@ -205,7 +238,8 @@ namespace pkg_rr {
             return res.get_future();
         }
         else {
-            msg() << "Checking for rebuild-requested installed packages (rebuild=YES)" << std::endl;
+            env.msg() << "Checking for rebuild-requested installed packages "
+                      << _var_sty("(rebuild=YES)") << std::endl;
             return scanner.add_axis("rebuild");
         }
     }
@@ -218,14 +252,16 @@ namespace pkg_rr {
             return res.get_future();
         }
         else {
-            msg() << "Checking for unsafe installed packages (" << UNSAFE_VAR << "=YES)" << std::endl;
+            env.msg() << "Checking for unsafe installed packages "
+                      << _var_sty('('_ch << UNSAFE_VAR << "=YES)") << std::endl;
             return scanner.add_axis(UNSAFE_VAR);
         }
     }
 
     void
     rolling_replacer::recheck_unsafe(pkgxx::pkgbase const& base) {
-        msg() << "Re-checking for unsafe installed packages (" << UNSAFE_VAR << "=YES)" << std::endl;
+        env.msg() << "Re-checking for unsafe installed packages "
+                  << _var_sty('('_ch << UNSAFE_VAR << "=YES)") << std::endl;
         auto const& PKG_INFO = env.PKG_INFO.get();
         pkgxx::guarded<todo_type> unsafe_pkgs;
         {
@@ -248,7 +284,7 @@ namespace pkg_rr {
                     // to replicate the logic just for our dry-run.
                     n.start_soon(
                         [&, unsafe_pkg = std::move(unsafe_pkg)]() {
-                            auto build_info  = pkgxx::build_info(PKG_INFO, unsafe_pkg.base);
+                            auto build_info  = pkgxx::build_info(PKG_INFO, unsafe_pkg);
                             auto unsafe_path = build_info.find("PKGPATH");
                             assert(unsafe_path != build_info.end());
                             unsafe_pkgs.lock()->emplace(unsafe_pkg.base, unsafe_path->second);
@@ -258,7 +294,7 @@ namespace pkg_rr {
                     n.start_soon(
                         [&, unsafe_pkg = std::move(unsafe_pkg)]() {
                             std::optional<pkgxx::pkgpath> unsafe_path;
-                            for (auto const& [var, value]: pkgxx::build_info(PKG_INFO, unsafe_pkg.base)) {
+                            for (auto const& [var, value]: pkgxx::build_info(PKG_INFO, unsafe_pkg)) {
                                 if (var == "PKGPATH") {
                                     unsafe_path.emplace(value);
                                 }
@@ -302,32 +338,27 @@ namespace pkg_rr {
 
     void
     rolling_replacer::dump_todo() const {
+        auto out = env.verbose();
+
         if (opts.just_fetch) {
-            atomic_verbose(
-                opts,
-                [&](auto& out) {
-                    out << "Packages to fetch:" << std::endl;
-                    dump_todo(out, "MISMATCH_TODO", MISMATCH_TODO);
-                    dump_todo(out, "MISSING_TODO" , MISSING_TODO );
-                });
+            out << "Packages to fetch:" << std::endl;
+            dump_todo(out, "MISMATCH_TODO", MISMATCH_TODO);
+            dump_todo(out, "MISSING_TODO" , MISSING_TODO );
         }
         else {
-            atomic_verbose(
-                opts,
-                [&](auto& out) {
-                    out << "Packages to rebuild:" << std::endl;
-                    dump_todo(out, "MISMATCH_TODO", MISMATCH_TODO);
-                    dump_todo(out, "REBUILD_TODO" , REBUILD_TODO );
-                    dump_todo(out, "MISSING_TODO" , MISSING_TODO );
-                    dump_todo(out, "UNSAFE_TODO"  , UNSAFE_TODO  );
-                });
+            out << "Packages to rebuild:" << std::endl;
+            dump_todo(out, "MISMATCH_TODO", MISMATCH_TODO);
+            dump_todo(out, "REBUILD_TODO" , REBUILD_TODO );
+            dump_todo(out, "MISSING_TODO" , MISSING_TODO );
+            dump_todo(out, "UNSAFE_TODO"  , UNSAFE_TODO  );
         }
-        vsleep(opts, 2s);
+
+        env.vsleep(2s);
     }
 
     void
     rolling_replacer::dump_todo(
-        std::ostream& out, std::string const& label, todo_type const& todo) const {
+        pkgxx::ttystream_base& out, std::string const& label, todo_type const& todo) const {
 
         // We could simply list packages in the alphabetical order, but
         // here we tsort them so that packages that are going to be
@@ -344,7 +375,7 @@ namespace pkg_rr {
                     else {
                         out << ' ';
                     }
-                    out << it->first;
+                    out << (num_pkgs % 2 == 0 ? _even_sty : _odd_sty)(it->first);
                     num_pkgs++;
                 }
             }
@@ -361,7 +392,7 @@ namespace pkg_rr {
         }
         out << ']';
         if (num_pkgs > 0) {
-            out << " (" << num_pkgs << ' '
+            out << " (" << tty::dull_colour(tty::yellow)(num_pkgs) << ' '
                 << (num_pkgs == 1 ? "package" : "packages")
                 << ')';
         }
@@ -390,7 +421,7 @@ namespace pkg_rr {
 
     pkgxx::graph<pkgxx::pkgbase, void, true>
     rolling_replacer::depgraph_installed() const {
-        msg() << "Building dependency graph for installed packages" << std::endl;
+        env.msg() << "Building dependency graph for installed packages" << std::endl;
 
         // There are two ways to build it. First, enumerate all the
         // installed packages and see which packages they depend
@@ -479,7 +510,7 @@ namespace pkg_rr {
 
     pkgxx::pkgversion
     rolling_replacer::update_depends_with_source(pkgxx::pkgbase const& base, pkgxx::pkgpath const& path) {
-        msg() << "Checking if " << base << " has new depends..." << std::endl;
+        env.msg() << "Checking if " << _pkgname_sty(base) << " has new depends..." << std::endl;
         auto const old_depends = topology.out_edges(base).value();
         auto const source      = source_depends(base, path);
         auto const& [version, new_depends] = source;
@@ -545,24 +576,24 @@ namespace pkg_rr {
             > const& old_depends,
         std::map<pkgxx::pkgbase, pkgxx::pkgpath> const& new_depends) {
 
-        auto out = msg();
+        auto out = env.msg();
         bool is_first = true;
         for (auto const& [dep_base, _pair]: new_depends) {
             if (old_depends.count(dep_base) == 0) {
                 if (is_first) {
-                    out << base << " has the following new depends (need to re-tsort):" << std::endl
+                    out << _pkgname_sty(base) << " has the following new depends (need to re-tsort):" << std::endl
                         << '[';
                     is_first = false;
                 }
                 else {
                     out << ' ';
                 }
-                out << dep_base;
+                out << _new_deps_sty(dep_base);
             }
         }
         if (!is_first) {
             out << ']' << std::endl;
-            vsleep(opts, 2s);
+            env.vsleep(2s);
         }
     }
 
@@ -601,7 +632,8 @@ namespace pkg_rr {
         }
 
         if (opts.dry_run) {
-            msg() << "Would run: " << pkgxx::stringify_argv(argv) << std::endl;
+            env.msg() << tty::faint("Would run: "_ch << pkgxx::stringify_argv(argv))
+                      << std::endl;
         }
         else if (opts.log_dir) {
             auto const version  = DEPENDS_CHECKED.find(base);
@@ -713,7 +745,7 @@ namespace pkg_rr {
                         continue;
                     }
                 }
-                warn() << "Invalid dependency: `" << dep << "' in " << var << std::endl;
+                env.warn() << "Invalid dependency: `" << dep << "' in " << var << std::endl;
             }
         }
 
@@ -779,7 +811,7 @@ namespace pkg_rr {
 
     void
     rolling_replacer::fetch(pkgxx::pkgbase const& base, pkgxx::pkgpath const& path) {
-        msg() << "Fetching " << base << std::endl;
+        env.msg() << "Fetching " << base << std::endl;
         run_make(base, path, {"fetch", "depends-fetch"}, make_vars_for_pkg(base));
     }
 
@@ -789,10 +821,10 @@ namespace pkg_rr {
 
         bool const was_installed = is_pkg_installed(base);
         if (was_installed) {
-            msg() << "Replacing " << base << std::endl;
+            env.msg() << "Replacing " << _pkgname_sty(base) << std::endl;
         }
         else {
-            msg() << "Installing " << base << std::endl;
+            env.msg() << "Installing " << _pkgname_sty(base) << std::endl;
         }
 
         auto make_vars = make_vars_for_pkg(base);
@@ -858,59 +890,25 @@ namespace pkg_rr {
     void
     rolling_replacer::clean(pkgxx::pkgbase const& base,
                             pkgxx::pkgpath const& path) {
-#if ENABLE_FAST_CLEAN
-        // "make clean" is slow because invoking "make" is slow. Do what it
-        // does, but in C++.
-        msg() << "Cleaning " << base << std::endl;
 
-        // The definition of BUILD_DIR changes depending on whether
-        // WRKOBJDIR is defined. See mk/bsd.prefs.mk
-        auto const& PKGSRCDIR       = env.PKGSRCDIR.get();
-        auto const& WRKOBJDIR       = env.WRKOBJDIR.get();
-        auto const& WRKDIR_BASENAME = env.WRKDIR_BASENAME.get();
-        auto const& BUILD_DIR
-            = WRKOBJDIR.empty() ? PKGSRCDIR / path : WRKOBJDIR / path;
-
-        // mk/bsd.prefs.mk defines WRKDIR with "?=". Hope nobody overrides
-        // it...
-        auto const& WRKDIR = BUILD_DIR / WRKDIR_BASENAME;
-
-        // Do what mk/bsd.pkg.clean.mk (su-do-clean) does.
-        try {
-            fs::remove_all(WRKDIR);
-            if (!WRKOBJDIR.empty()) {
-                // This directory is supposed to be empty.
-                fs::remove(BUILD_DIR);
-
-                // Ignore errors from rmdir(2). This directory can
-                // legitimately be non-empty.
-                std::error_code ec;
-                fs::remove(BUILD_DIR.parent_path(), ec);
-
-                // When ${CREATE_WRKDIR_SYMLINK} is set to YES, a symlink
-                // to WRKDIR is created. Remove it.
-                fs::remove(PKGSRCDIR / path / WRKDIR_BASENAME);
-            }
-        }
-        catch (fs::filesystem_error&) {
-            // But this will fail when WRKDIR has non-writable
-            // directories. Fall back to "make clean" when that happens. It
-            // should now how to handle it.
-            run_make(base, path, {"clean"}, opts.make_vars);
-        }
-#else
+        // "make clean" is slow because invoking "make" is slow. It's
+        // tempting to simulate what it does in C++. However, doing it
+        // properly is suprisingly hard because there are so many
+        // controlling knobs (e.g. WRKOBJDIR, WRKDIR_BASENAME, ...) so
+        // don't even think of that. We have tried in the past, and failed
+        // twice. Also don't forget that packages may define
+        // {pre,post}-clean targets.
         run_make(base, path, {"clean"}, opts.make_vars);
-#endif
     }
 
     void
     rolling_replacer::report() const {
         if (opts.verbose > 0) {
             for (auto const& base: SUCCEEDED) {
-                std::cout << "+ " << base << std::endl;
+                env.raw_msg() << tty::dull_colour(tty::green)("+ "_ch << base) << std::endl;
             }
             for (auto const& base: FAILED) {
-                std::cout << "- " << base << std::endl;
+                env.raw_msg() << tty::colour(tty::red)("- "_ch << base) << std::endl;
             }
         }
     }
